:- ensure_loaded( '../pe/pe' ).	% fam/1.
	% also
	% sel_option_actions/4, sload/1, slp_file_location/2, sls/1.
	% experiment/1, exper_setrand/1.
:- ensure_loaded( 'src/lib/new_directory' ).			% /1.
:- ensure_loaded( 'src/lib/start_billow_recording' ).	% /5.
:- ensure_loaded( 'src/lib/datime_to_readable' ).		% /4.
:- ensure_loaded( 'src/lib/werr' ).				% /1.
:- ensure_loaded( 'src/lib/secure_dir' ).			% /2.
:- ensure_loaded( 'src/lib/id_dirname' ).			% /2.
:- ensure_loaded( 'src/lib/clean_module' ).			% /1.
:- ensure_loaded( 'src/lib/atoms_concat' ).			% /2.
:- ensure_loaded( 'src/lib/delete_file_if' ).		% /1.
:- ensure_loaded( 'src/pe_multi' ).				% /N.
:- ensure_loaded( 'src/plot_multi_trips' ).			% /4.
:- ensure_loaded( library(random) ).				% get/setrand/1.
:- ensure_loaded( library(lists) ).				
	% member/2,memberchk/2,nth/3.
:- ensure_loaded( library(system) ).
	% system/1,working_directory/2,datime/1.
:- bb_put( billow_version, 0:0:2 ).

% TO DO :
%       : FILE FOR REDIRECTING FAM OUTPUT (FAM OPTION ?)
%
billow( Opts ) :-
	% add random seeds support
	sel_option_actions( slp(Slp), Opts, true, required(slp) ),
	sel_option_actions( dir(Dir), Opts, true, billow_ddir(Dir) ),
	new_directory( Dir, true, abort ),
	% copy_slp( Slp, Dir, CpSlp ),
	sel_option_actions( goal(AtGoal), Opts, true, required(goal) ),
	parse_atomic_goal( AtGoal, Goal ),
	ToPreG = billow_load_in( exp_pregoal, PreGF ),
	Null   = null(true),
	sel_option_actions( pregoal(PreGF), Opts, ToPreG, PreGF=Null ),
	sel_option_actions( datafile(File), Opts, consult(File), required(datafile) ),
	sel_option_actions( ids(Ids), Opts, true, true ),
	ToPrr = billow_load_in( exp_prior, PrrF ),
	sel_option_actions( prior(PrrF), Opts, ToPrr, PrrF=random ),
	LdPdY = (billow_load_in(exp_predictive_yields,PYsFl),PdYs=true,true_yield_prbs(Slp)),
	PdYMs = [['Not option predictive_yields/1 provided.'],
		    ['Switching predictive statistics off.']],
	PdYNo = (werr(PdYMs),PdYs=Null),
	ToPdY = sel_option_actions( predictive_yields(PYsFl), Opts, LdPdY, PdYNo ),
	sel_option_actions( predictive_rms(false), Opts, PdYs=Null, ToPdY ),
	sel_option_actions( seeds(Seeds), Opts, true, Seeds=auto ),
	optrand( Seeds ),
	sel_option_actions( info(Info), Opts, true, Info='Readme.txt' ),
	sel_option_actions( fam(FamInOpts), Opts, true, FamInOpts=[] ),
	% working_directory( OldWd, Dir ), 
	batch_has_experiments( Ids, Exprs ),
	start_billow_recording( Dir, Slp, Info, Exprs, Opts, Out, OldOut, Err ),
	FamConOpts = [slp-Slp,pregoal-PreGF,goal-Goal,prior-PrrF,dir-Dir],
	BilConOpts = [predictive_rms-PdYs],
	cohese_fam_options( FamConOpts, FamInOpts, FamOpts ),
	cohese_fam_options( BilConOpts, [], BilOpts ),
	Run = run_experiments( Exprs, BilOpts, FamOpts ),
	catch( Run, Any, run_experiments_excp(Any) ),
	write( 'End of batch.' ), nl,
	datime( Datime ),
	datime_to_readable( Datime, ReadableCs ),
	atom_codes( Readable, ReadableCs ),
	write( Readable ), write( '.' ), nl,
	!,
	set_output( OldOut ),
	set_prolog_flag( user_error, Err ),
	sel_option_actions( secure(Sec), Opts, secure_dir(Sec,Dir), true ),
	close( Out ),
	% the following is also called at exception end
	sel_option_actions( at_end(End), Opts, call(End), true ).
	% working_directory( _DirAgain, OldWd ). 

run_experiments( [], _BilOpts, _FamOpts ) :-
	delete_file_if( '.idatafile' ),
	delete_file_if( '.foofoutos' ).
run_experiments( [expr_data(Id,Rs,Ds,Ls)|T], BilOpts, FamOpts ) :-
	% Here
	record_exp_start( Id, Rs, Ds, Ls ),
	id_dirname( Id, FamOpts, Dir ),
	% pe_multi( Runs, ppway6, Dir, Titles, MultiTrips ),
	% atom_concat( MotherDir, IdDir, Dir ),
	pe_multi( Rs, BilOpts, FamOpts, Dir, Ls, MultiTrips ),
	plot_multi_trips( MultiTrips, Dir, Ds, Ls ),
	run_experiments( T, BilOpts, FamOpts ).

run_experiments_excp( Any ) :-
	write( 'Caught the following while running batch.' ), nl,
	print_message( error, Any ).

batch_has_experiments( Ids, Exprs ) :-
	( current_predicate(expr_data/4) ->
		Goal = expr_data(Id,_Runs,_Dims,_Dirs),
		findall( Goal, (member(Id,Ids),call(Goal)), Exprs )
		;
		Err = [['Cannot find definition for expr_data/4.'],
		       ['This predicate defines the kinds of experiments to perform.'],
			  ['The file containing definition of expr_data, should be the one provided with the datafile option']
			 ],
		werr( Err ), abort
	).

required( OptId ) :- 
	Mess = [[ 'Cannot continue, because option ', OptId,
	          ' although required, is missing.' ]],
	werr( Mess ),
	abort.

parse_atomic_goal( PreGoal, Goal ) :-
	( (atomic(PreGoal),\+ atom_concat(_SGoal,'.',PreGoal)) ->
		atom_concat( PreGoal, '.', AtGoal )
		;
		AtGoal = PreGoal
	),
	WR = ( write_to_chars( AtGoal, Cs ), 
	       read_from_chars( Cs, Goal )
		),
	( atomic(AtGoal) ->
		( catch(WR,_Any,parse_atomic_goal_exc(AtGoal)) ->
			true
			;
			parse_atomic_goal_exc( AtGoal )
		)
		;
		Goal = AtGoal
	).

parse_atomic_goal_exc( AtGoal ) :-
		Mess = [['Cannot parse option goal, ',AtGoal],
			   ['Please ensure it is in form parsable by read/1.'],
			   ['This should include the terminating period.']],
		werr( Mess ),
		abort.

record_exp_start( I, R, M, D ) :-
	write( starting(I,R,M,D) ), nl,
	datime( Datime ),
	datime_to_readable( Datime, RdCs ),
	atom_codes( Rd, RdCs ), 
	write( Rd ), nl.

billow_load_in( Module, InFile ) :-
	( atomic(InFile) -> 
		File = InFile 
		;
		InFile =.. [File|_]
	),
	clean_module( Module ),
	( (Module == exp_prior, (File == random; File == uniform)) ->
		true
		;
		Module:consult( File )
	).

true_yield_prbs( Slp ) :-
	sload( Slp ), 
	findall( Prb, 
			( exp_predictive_yields:yield(Goal),
			  scall_sum_refutations(Goal,Prb)
			),
				TruePrbs ),
	bb_put( true_yield_prbs, TruePrbs ).


optrand( Rand ) :-
	( Rand == auto ->
		exper_setrand( _NewRand )
		;
		setrand( Rand )
	).
	
rerand :- 
	getrand( Rand ),
	setrand( Rand ).

cohese_fam_options( [], FamOpts, FamOpts ).
cohese_fam_options( [Name-Value|T], InOpts, FamOpts ) :-
	( Value == null(true) ->
		MidOpts = InOpts
		;
		TestTerm =.. [Name,NewValue],
		( memberchk(TestTerm,InOpts) ->
			Mess = [ ['Multiple values for FAM option ', Name, '/1.'],
			         ['Both ', NewValue, ', and ', Value, ', found.']
			       ],
			werr( Mess ),
			abort
			;
			( cohese_fam_load_option_pad(Name,Value,NewValue) ->
				true
				;
				NewValue = Value
			),
			Term =.. [Name,NewValue],
			MidOpts = [Term|InOpts]
		)
	),
	cohese_fam_options( T, MidOpts, FamOpts ).

cohese_fam_load_option_pad( prior, Value, NewValue ) :-
	Value \= uniform, 
	Value \= random,
	atom_concat( 'exp_', prior, Module ),
	NewValue = Module:Value.
cohese_fam_load_option_pad( pregoal, Value, NewValue ) :-
	% atom_concat( 'exp_', pregoal, Module ),
	NewValue = exp_pregoal:Value.

copy_slp( Slp, Dir, BaseSlp ) :-
	slp_file_location( Slp, RelSlp ),
	( file_exists( RelSlp ) ->
		basename( RelSlp, BaseSlp ),
		( atom_concat( _DirNoSl, '/', Dir ) ->
			DirSl = Dir 
			;
			atom_concat( Dir, '/', DirSl )
		),
		atoms_concat( ['cp ',RelSlp,' ',DirSl,BaseSlp], Cp ),
		write( system(Cp) ), nl,
		system( Cp )
		;
		Mess = [['Slp, ',RelSlp,', does not exist.']],
		werr( Mess ),
		abort
	).

billow_ddir( Dir ) :-
	datime( datime(Year,Month,Day,Hour,Minute,_Sec) ),
	number_codes( Year, Yd ), 
	number_codes( Day, D ), two_digit_no( D, Dd ),
	number_codes( Hour, H ), two_digit_no( H, Hd ),
	number_codes( Minute, M ), two_digit_no( M, Md ),
	nth( Month, [aa,ab,ac,ba,bb,bc,ca,cb,cc,da,db,dc], Nth ),
	atom_codes( Nth, Nd ),
	append_list( [Yd,Nd,Dd,"_",Hd,Md], DirCs ),
	atom_codes( Dir, DirCs ).

two_digit_no( No, TwoDigs ) :-
	( No = [One] ->
		TwoDigs = [0'0,One]
		;
		No = TwoDigs
	).
