:- ensure_loaded( library('option/select_preferred') ).	% /3.
:- ensure_loaded( library(system) ).		
	% file_exists/1, file_property/2, make_directory/1.
	% working_directory/1, rename_file/2.
:- ensure_loaded( library(lists) ).		% memberchk/2.	
:- ensure_loaded( library('term/flatoms') ).	% /3.
:- ensure_loaded( library('file/unique_filename') ).	% /2.
:- ensure_loaded( library('file/fname_extension') ).	% /5.

gpr_defaults(	[	
		ids(_I),dir(gpr),repo('repo'),
		out(gnp_out),x(_X),y(_Y),
		xl(no__label),yl(no__label),
		keep(eps) ]).  

gpr( InArgs ) :-
	collect_option( id, ids, InArgs, Args ),
	( memberchk(help,Args) ->
		write( user_error, 'gpr: gnuplot from repository file. Matches terms of a repository terms (.pl) file, and generates dir, data files and gnucom file.' ),
		nl( user_error ),
		nl( user_error ),
		write( user_error, '  repo(Repo):  .pl repository file, *repo*.' ),
		nl( user_error ),
		write( user_error, '      id(Id):  multiple Ids to match.' ),
		nl( user_error ),
		write( user_error, '    out(Out):  stem name for gnu coms and eps.' ),
		nl( user_error ),
		write( user_error, '    dir(Dir):  dir to create, *gpr*.' ),
		nl( user_error ),
		write( user_error, '        x(X):  x range *auto*.' ),
		nl( user_error ),
		write( user_error, '        y(Y):  y range *auto*.' ),
		nl( user_error ),
		write( user_error, '      xl(XL):  x label (*no label*).' ),
		nl( user_error ),
		write( user_error, '      yl(YL):  y label (*no label*).' ),
		nl( user_error ),
		write( user_error, '     keep(K):  {all,*eps*}.' ),
		nl( user_error ),
		nl( user_error )
		;
		gpr_defaults( Defs ),
		select_preferred( ids(Ids), Args, Defs ),
		select_preferred( dir(Dir), Args, Defs ),
		select_preferred( repo(PrvRepo), Args, Defs ),
		( file_exists(PrvRepo) ->
			Repo = PrvRepo
			;
			fname_extension( PrvRepo, ".pl", _St, Repo, _Fl ),
			atom_concat( PrvRepo, '.pl', Repo )
		),
		select_preferred( out(Out), Args, Defs ),
		select_preferred( x(X), Args, Defs ),
		select_preferred( y(Y), Args, Defs ),
		unique_filename( Dir, UqDir ),
		make_directory( UqDir ),
		flatoms( ['cp ',Repo,' ',UqDir,'/repo.pl'], Cp ), 
		system( Cp ),
		working_directory( Old, UqDir ),
		select_preferred( xl(XL), Args, Defs ),
		select_preferred( yl(YL), Args, Defs ),
		gp_rps_ids( Ids, Out, X, Y, XL, YL ),
		flatoms( ['gnuplot ',Out,'.gpc'], GnuP ),
		system( GnuP ),
		select_preferred(keep(Keep),Args,Defs),
		( Keep == eps ->
			atom_concat( Out, '.eps', Eps ),
			atom_concat( '../', Eps, UpEps ),
			rename_file( Eps, UpEps ),
			system( 'rm -rf *' ),
			working_directory( _UD, Old ),
			delete_file( UqDir )
			;
			working_directory( _UD, Old )
		)
	).
	
gp_rps_ids( Ids, Out, Xa, Ya, XL, YL ) :-
	consult( repo ),
	findall( Id-Data, stats(Id,Data), IData ),
	% initial_ranges( IData, InX, InY ),
	gp_rps_matching( IData, Ids, Match, _InX, _InY, Xi, Yi ),
	( var(Xa) -> X = Xi ; X = Xa ),
	( var(Ya) -> Y = Yi ; Y = Ya ),
	to_eps( Match, X, Y, Out, XL, YL ).
	% flatoms( ['mv ',Out,' ',Old], Mv ),
	% system( Mv ).

% initial_ranges( IData, InX, InY ) :-

gp_rps_matching( [], _Ids, [], X, Y, X, Y ).
gp_rps_matching( [Hid-Hdt|T], Ids, Match, InX, InY, OutX, OutY ) :-
	( memberchk(Hid,Ids) ->
		atom_concat( Hid, '.dat', Mid ),
		tell( Mid ),
		data_to_output( Hdt, InX, InY, MdX, MdY ),
		told,
		Match = [Hid-Mid|Tatch]
		;
		Match = Tatch
	),
	gp_rps_matching( T, Ids, Tatch, MdX, MdY, OutX, OutY ).

data_to_output( [], IXl:IXh, IYl:IYh, Xl:Xh, Yl:Yh ) :-
	gnu_range( IXl, IXh, Xl, Xh ),
	gnu_range( IYl, IYh, Yl, Yh ).

data_to_output( [A-B-C|T], InX, InY, OutX, OutY ) :-
	write( A ),
	write( ' ' ),
	write( B ),
	write( ' ' ), 
	write( C ), 
	nl, 
	( var(InX) ->
		MdX = A:A
		;
		range_expansion( A, InX, MdX )
	),
	( var(InY) ->
		Half is C / 2,
		MinY is B - Half,
		MaxY is B + Half,
		MdY = MinY:MaxY
		;
		range_expansion( B, C, InY, MdY )
	),
	data_to_output( T, MdX, MdY, OutX, OutY ).

gnu_range( Low, High, Lower, Higher ) :-
	Margin is (Low + High) / 20,
	Lower is Low - Margin,
	Higher is High + Margin.

range_expansion( Val, Low:Up, NwL:NwUp ) :-
	NwL is min(Val,Low),
	NwUp is max(Val,Up).

range_expansion( Val, Err, Low:Up, NwL:NwUp ) :-
	Half is Err / 2,
	MinVal is Val - Half,
	MaxVal is Val + Half,
	NwL is min(MinVal,Low),
	NwUp is max(MaxVal,Up).

to_eps( Match, X, Y, Out, XL, YL ) :-
	atom_concat( Out, '.gpc', Gpc ),
	atom_concat( Out, '.eps', Eps ),
	tell( Gpc ),
	write( 'set terminal postscript eps colour' ), nl,
	write( 'set output "' ), 
	write( Eps ), write( '"' ), nl,
	write_eps_label( XL, x ),
	write_eps_label( YL, y ),
	write( 'set multiplot' ), nl,
	write( 'set xrange [' ), write( X ), write( ']' ), nl,
	write( 'set yrange [' ), write( Y ), write( ']' ), nl,
	write( 'plot ' ),
	write_match_plot( Match ),
	nl,
	told.

write_eps_label( Lab, W ) :-
	( Lab == no__label ->
		true
		;
		write( 'set ' ), write( W ), write( 'label "' ),
		write( Lab ), write( '"' ), nl
	).

write_match_plot( [Tit-Dat|T] ) :-
	write( '"' ), write( Dat ), write( '" using 1:2 title "' ),
	write( Tit ), write( '" with lines, "' ), write( Dat ),
	write( '" notitle with yerrorbars' ), 
	( T = [] ->
		true
		;
		write( ', '),
		write_match_plot( T )
	).

% gp_rps_matching( [Hid-Hdt|T], Ids, Match ),
	% to_gnufile( 

collect_option( Name, Collective, InArgs, Args ) :-
	collect_option_1( InArgs, OutArgs, Name, List ),
	( List = [] ->
		Args = OutArgs
		;
		ColT =.. [Collective,List],
		Args = [ColT|OutArgs]
	).

collect_option_1( [], [], _, [] ).
collect_option_1( [H|T], Args, Name, List ) :-
	Term =.. [Name,Para],
	( Term = H -> 
		List = [Para|TList],
		Args = TArgs
		;
		List = TList,
		Args = [H|TArgs]
	),
	collect_option_1( T, TArgs, Name, TList ).
