:- ensure_loaded( 'lib/werr' ).			% /1.
:- ensure_loaded( library(system) ).		
	% file_exists/1, make_directory/1.
:- ensure_loaded( library(lists) ). 		% append/3.		
:- ensure_loaded( 'lib/sample_that_many' ).		
:- ensure_loaded( 'lib/mean_and_emp_std_dev_list' ).
:- ensure_loaded( 'lib/append_list' ).				% /2.
:- ensure_loaded( 'lib/root_mean_square_error' ).		% /3.

pe_multi( [], _BilOpts, _FamOpts, _BDir, _Dirs, [] ).
pe_multi( [Htms-Hpoints|T], BilOpts, FamOpts, BDir, [Hd|Td], [Htrip|Ttrip] ) :-
	% atom_concat( 'v7_', Hd, Vhd ),
	atom_concat( BDir, Hd, FDir ),
	atom_codes( FDir, FdCs ),
	append( FdCs, [0'/], Dir ),
	atom_codes( DirAt, Dir ),
	( file_exists( DirAt ) ->
		werr( [['Directory exists: ',DirAt]] ), abort
		;
		make_directory( DirAt )
	),
	pe_iter( Hpoints, Htms, BilOpts, FamOpts, Dir, Htrip ),
	!,
	pe_multi( T, BilOpts, FamOpts, BDir, Td, Ttrip ).

pe_iter( [], _Tms, _BOpts, _FOpts, _Dir, [] ).
pe_iter( [H|T], Tms, BOpts, FOpts, Dir, [H-HMean-HSD|TTrips] ) :-
	write( multi_sampling_at(H) ), nl,
	% sload( Slp ),
	% bb_get( pp, TruePPs ),
	% kv_decompose( TruePPs, _Keys, TruePPvs ),
	( memberchk(predictive_rms(true),BOpts) ->
		true
		;
		Prmss = []
	),
	pe_tms_iter( 1, Tms, H, FOpts, Dir, RMSs, Prmss, Its, RTms ),
	mean_and_emp_std_dev_list( RMSs, HMean, HSD ),
	store_rmsms_iters_rtms( Dir, H, HMean, HSD, Its, RTms ),
	( Prmss == [] ->
		true
		;
		mean_and_emp_std_dev_list( Prmss, PdvMean, PdvSD ),
		store_predictive_rmss( Dir, H, PdvMean, PdvSD )
	),
	% write( mean_and_std_dev_list( SSDs, HMean, HSD ) ), nl,
	% mean_and_std_dev_list( SSDs, HMean, HSD ),
	% write( mean_and_emp_std_dev_list( SSDs, HMean, HSD ) ), nl,
	!,
	pe_iter( T, Tms, BOpts, FOpts, Dir, TTrips ).

pe_tms_iter( I, Tms, _Samples, _FOpts, _Dir, Rxtis, Prmss, Its, RTms ) :-
	I > Tms,
	!,
	Rxtis = [],
	Prmss = [],
	Its = [],
	RTms = [].
pe_tms_iter( I, Tms, Samples, FOpts, Dir, Rxtis, Prmss, Its, RTms ) :-
	write( repetition(I) ),
	gen_data( Samples, FOpts, '.idatafile', TruePPs ),
	write( '_gen_' ),
	% ( (Samples =:= 310,I =:=5) -> trace ; true ),
	Tmp = '.idatafile',
	pe_main( FOpts, Tmp, PPs, It, RTm ),
	write( '_main_' ),
	kv_decompose( PPs, _Keys, PPvs ),
	write( dec_ ),
	root_mean_square_error( TruePPs, PPvs, Rxti ),
	predictive_rms( Prmss, Tprms ),
	write( sum ), nl,
	Rxtis = [Rxti|TRxtis],
	Its = [It|Tits],
	RTms = [RTm|RTtms],
	NxI is I + 1,
	!,
	pe_tms_iter( NxI, Tms, Samples, FOpts, Dir, TRxtis, Tprms, Tits, RTtms ).

predictive_rms( Prmss, Tprms ) :-
	( Prmss == [] ->
		Tprms = []
		;
		findall( Prb,
			( exp_predictive_yields:yield(Goal),
			  scall_sum_refutations(Goal,Prb)
			), Prbs ),
		bb_get( true_yield_prbs, TruePrbs ),
		root_mean_square_error( TruePrbs, Prbs, Hrms ),
		Prmss = [Hrms|Tprms]
	).


pe_main( InOpts, DataFile, PPs, I, Tm ) :-
	% ppway_pe_data( Data ),
	% ubiquitous_metabolites( Metbls ),
	% open_null_stream( Null ),
	% current_output( OldOut ),
	% set_output( Null ),
	write( streams_ ),
	statistics( runtime, [_St1,_Ls1] ),
	sel_option_actions( termin(_Cond), InOpts, Opts=InOpts, Opts=[termin( [ll_e(1.0e-02)] )|InOpts] ),
	% ReGoal = (set_output(OldOut),pe_main(Slp,DataFile,PPs)),
	FamOpts = [return( [iter(I),termin(Why),final_pps(PPs)] ),
			 datafile(DataFile),
			 count(exact),
			 writeps(last)|
			 Opts
	          ],
	fam( FamOpts ),
	statistics( runtime, [_St2,Tm] ),
	write( i-I ), write( '_with_' ), write( termin(Why) ).
	% set_output( OldOut ),
	% close( Null ).

gen_data( ThatMany, FamOpts, File, TruePPs ) :-
	memberchk( slp(Slp), FamOpts ),
	memberchk( pregoal(PGl), FamOpts ),
	call( PGl ),
	memberchk( goal(Goal), FamOpts ),
	sample_that_many( Slp, ThatMany, Goal, Freq ),
	current_output( OldOut ),
	open( File, write, Stream ),
	set_output( Stream ),
	write( frequencies(Freq) ),
	write( '.' ), nl,
	set_output( OldOut ),
	close( Stream ),
	bb_get( pp, TruePPvs ),
	kv_decompose( TruePPvs, _Keys, TruePPs ).

store_one_iteration( DirCs, Samples, Iter, Mean, S2 ) :-
	number_codes( Samples, S ),
	number_codes( Iter, I ),
	append_list( [DirCs,"s_",S,"_i_",I], FnameCs ),
	atom_codes( Fname, FnameCs ),
	open( Fname, write, Stream ),
	write( Stream, Mean ),
	write( Stream, '  ' ),
	write( Stream, S2 ),
	nl( Stream ),
	close( Stream ).

store_rmsms_iters_rtms( Dir, Smp, Mean, SDv, Its, RTms ) :-
	append( Dir, "rmse_mean_sdv", MSdvFile ),
	append( Dir, "iterations", ItsFile ),
	append( Dir, "runtime", RtmFile ),
	term_to_file_cs( mean_sdv(Smp,Mean,SDv), MSdvFile ),
	term_to_file_cs( its(Smp,Its), ItsFile ),
	term_to_file_cs( rtms(Smp,RTms), RtmFile ).

store_predictive_rmss( Dir, Smp, Mean, SDv ) :-
	append( Dir, "predictive_rms_mean_n_sdv", File ),
	term_to_file_cs( mean_sdv(Smp,Mean,SDv), File ).

term_to_file_cs( Term, FileCs ) :-
	atom_codes( File, FileCs ),
	open( File, append, S ),
	write( S, Term ),
	write( S, '.' ),
	nl( S ),
	close( S ).
